"""Tasks for Discord Notify."""

import random

from celery import shared_task
from discordproxy.exceptions import DiscordProxyTimeoutError

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .core import forward_notification_to_discord

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@shared_task(bind=True)
def task_forward_notification_to_discord(
    self,
    notification_id: int,
    discord_uid: int,
    title: str,
    message: str,
    level: str,
    timestamp: str,
):
    """Forward a notification to discord as task."""
    try:
        forward_notification_to_discord(
            notification_id=notification_id,
            discord_uid=discord_uid,
            title=title,
            message=message,
            level=level,
            timestamp=timestamp,
        )
    except DiscordProxyTimeoutError as ex:
        countdown = 30 + int(random.uniform(0, 5)) * 60
        logger.warning(
            "Timeout exceeded when trying to send notification %d to Discord. "
            "Trying again in %d seconds.",
            notification_id,
            countdown,
        )
        raise self.retry(countdown=countdown) from ex

    logger.info("Forwarded notification %d to Discord", notification_id)



"""Signals for Discord Notify."""

from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_save
from django.dispatch import receiver

from allianceauth.notifications.models import Notification
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import DISCORDNOTIFY_ENABLED, DISCORDNOTIFY_SUPERUSER_ONLY
from .tasks import task_forward_notification_to_discord

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@receiver(post_save, sender=Notification)
def forward_new_notifications(instance, created, **kwargs):
    """Forward a new notification received by signal."""
    if DISCORDNOTIFY_ENABLED:
        if created and (not DISCORDNOTIFY_SUPERUSER_ONLY or instance.user.is_superuser):
            logger.info(
                "Processing notification %d for: %s", instance.id, instance.user
            )
            try:
                discord_uid = instance.user.discord.uid
            except (AttributeError, ObjectDoesNotExist):
                logger.warning(
                    "Can not forward notification to user %s, "
                    "because he has no Discord account",
                    instance.user,
                )
                return
            # we are passing through the instance attributes,
            # because it is not guaranteed that the object has already been saved
            task_forward_notification_to_discord.delay(
                notification_id=instance.id,
                discord_uid=discord_uid,
                title=instance.title,
                message=instance.message,
                level=instance.level,
                timestamp=instance.timestamp.isoformat(),
            )
        else:
            logger.debug("Ignoring notification %d for: %s", instance.id, instance.user)



"""Views for Discord Notify."""

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect

from allianceauth.notifications import notify


@login_required
def send_test_notification(request):
    """Render view for sending a test notification."""
    notify(
        request.user,
        title="Test Notification",
        message=f"This is a test notification from Discord Notify created for {request.user}.",
    )
    messages.success(
        request, f"Discord Notify: Test notification was created for {request.user}"
    )
    return redirect("authentication:dashboard")



from unittest.mock import patch

from django.test import TestCase
from django.utils.timezone import now

from discordnotify.core import forward_notification_to_discord

MODULE = "discordnotify.core"


@patch(MODULE + ".DiscordClient", spec=True)
class TestForwardNotificationToDiscord(TestCase):
    def test_should_have_correct_proxy_settings(self, mock_DiscordClient):
        cases = [
            ("localhost", "50051", "localhost:50051"),
            ("127.0.0.1", "50051", "127.0.0.1:50051"),
            ("127.0.0.1", "50000", "127.0.0.1:50000"),
        ]
        for host, port, expected in cases:
            with self.subTest(f"{host}, {port} = {expected}"):
                with patch(MODULE + ".DISCORDPROXY_HOST", host), patch(
                    MODULE + ".DISCORDPROXY_PORT", port
                ):
                    forward_notification_to_discord(
                        notification_id=1,
                        discord_uid=2,
                        title="title",
                        message="message",
                        level="info",
                        timestamp=now().isoformat(),
                    )
                _, kwargs = mock_DiscordClient.call_args
                target = kwargs["target"]
                self.assertEqual(target, expected)



"""Forward Alliance Auth notifications to users on Discord."""

# pylint: disable = invalid-name
default_app_config = "discordnotify.apps.DiscordnotifyConfig"

__title__ = "Discord Notify"
__version__ = "2.0.0"



from unittest.mock import patch

import grpc
from celery.exceptions import Retry
from discordproxy.exceptions import DiscordProxyException, DiscordProxyTimeoutError

from django.contrib.auth.models import User
from django.db.models.signals import post_save
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse

from allianceauth.notifications import notify
from allianceauth.notifications.models import Notification
from allianceauth.services.modules.discord.models import DiscordUser
from app_utils.testing import create_fake_user

from discordnotify import views
from discordnotify.signals import forward_new_notifications
from discordnotify.tasks import task_forward_notification_to_discord

CORE_PATH = "discordnotify.core"
SIGNALS_PATH = "discordnotify.signals"
VIEWS_PATH = "discordnotify.views"
TASKS_PATH = "discordnotify.tasks"


@patch(CORE_PATH + ".DiscordClient.create_direct_message", spec=True)
@override_settings(CELERY_ALWAYS_EAGER=True)
class TestIntegration(TestCase):
    def setUp(self) -> None:
        self.user = create_fake_user(1001, "Bruce Wayne")

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_forward_when_new_notification_is_created(
        self, mock_create_direct_message
    ):
        # given
        DiscordUser.objects.create(user=self.user, uid=123)
        # when
        notify(self.user, title="title", message="message")
        # then
        self.assertTrue(mock_create_direct_message.called)
        _, kwargs = mock_create_direct_message.call_args
        self.assertEqual(kwargs["user_id"], 123)
        self.assertEqual(kwargs["embed"].description, "message")
        self.assertEqual(kwargs["embed"].title, "title")

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_split_messages_that_are_too_long(self, mock_create_direct_message):
        # given
        DiscordUser.objects.create(user=self.user, uid=123)
        # when
        notify(self.user, title="title", message="x" * 3000)
        # then
        self.assertTrue(mock_create_direct_message.called)
        _, kwargs = mock_create_direct_message.call_args
        self.assertEqual(len(kwargs["embed"].description), 2048)

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_not_forward_when_app_is_disabled(self, mock_create_direct_message):
        # given
        DiscordUser.objects.create(user=self.user, uid=123)
        # when
        with patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", False):
            notify(self.user, "hi")
        # then
        self.assertFalse(mock_create_direct_message.called)

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_not_forward_when_notification_is_updated(
        self, mock_create_direct_message
    ):
        # given
        DiscordUser.objects.create(user=self.user, uid=123)
        post_save.disconnect(forward_new_notifications, sender=Notification)
        notify(self.user, "hi")
        post_save.connect(forward_new_notifications, sender=Notification)
        # when
        notif = Notification.objects.filter(user=self.user).first()
        notif.mark_viewed()
        # then
        self.assertFalse(mock_create_direct_message.called)

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_not_forward_when_user_has_no_account(
        self, mock_create_direct_message
    ):
        # when
        notify(self.user, "hi")
        # then
        self.assertFalse(mock_create_direct_message.called)

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", True)
    def test_should_forward_to_superusers_only_1(self, mock_create_direct_message):
        # given
        DiscordUser.objects.create(user=self.user, uid=123)
        # when
        notify(self.user, "hi")
        # then
        self.assertFalse(mock_create_direct_message.called)

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", True)
    def test_should_forward_to_superusers_only_2(self, mock_create_direct_message):
        # given
        user = User.objects.create_superuser("Clark Kent")
        DiscordUser.objects.create(user=user, uid=987)
        # when
        notify(user, "hi")
        # then
        self.assertTrue(mock_create_direct_message.called)

    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", True)
    def test_should_forward_to_superusers_only_3(self, mock_create_direct_message):
        # given
        user = User.objects.create_superuser("Clark Kent")
        DiscordUser.objects.create(user=user, uid=987)
        post_save.disconnect(forward_new_notifications, sender=Notification)
        notify(user, "hi")
        post_save.connect(forward_new_notifications, sender=Notification)
        # when
        notif = Notification.objects.filter(user=user).first()
        notif.mark_viewed()
        # then
        self.assertFalse(mock_create_direct_message.called)

    @patch(CORE_PATH + ".DISCORDNOTIFY_MARK_AS_VIEWED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_mark_notification_as_viewed_once_submitted(
        self, mock_create_direct_message
    ):
        # given
        DiscordUser.objects.create(user=self.user, uid=123)
        # when
        obj = Notification.objects.notify_user(user=self.user, title="hi")
        # then
        obj.refresh_from_db()
        self.assertTrue(obj.viewed)

    @patch(CORE_PATH + ".DISCORDNOTIFY_MARK_AS_VIEWED", False)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_not_mark_notification_as_viewed_once_submitted(
        self, mock_create_direct_message
    ):
        # given
        DiscordUser.objects.create(user=self.user, uid=123)
        # when
        obj = Notification.objects.notify_user(user=self.user, title="hi")
        # then
        obj.refresh_from_db()
        self.assertFalse(obj.viewed)

    @patch(CORE_PATH + ".DISCORDNOTIFY_MARK_AS_VIEWED", False)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_ENABLED", True)
    @patch(SIGNALS_PATH + ".DISCORDNOTIFY_SUPERUSER_ONLY", False)
    def test_should_not_mark_notification_as_viewed_when_failed(
        self, mock_create_direct_message
    ):
        # given
        mock_create_direct_message.side_effect = DiscordProxyException
        DiscordUser.objects.create(user=self.user, uid=123)
        # when
        obj = Notification.objects.notify_user(user=self.user, title="hi")
        # then
        obj.refresh_from_db()
        self.assertFalse(obj.viewed)


class TestViews(TestCase):
    @patch(VIEWS_PATH + ".notify", wraps=notify)
    @patch(VIEWS_PATH + ".messages")
    def test_should_create_notification_and_send_message(
        self, spy_messages_plus, spy_notify
    ):
        # given
        user = User.objects.create_user("Bruce Wayne")
        factory = RequestFactory()
        request = factory.get(reverse("discordnotify:send_test_notification"))
        request.user = user
        # when
        response = views.send_test_notification(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("authentication:dashboard"))
        self.assertTrue(spy_notify.called)
        self.assertTrue(spy_messages_plus.success.called)


@patch(TASKS_PATH + ".logger", spec=True)
@patch(TASKS_PATH + ".forward_notification_to_discord", spec=True)
class TestTasks(TestCase):
    def test_should_send_notification(
        self, mock_forward_notification_to_discord, mock_logger
    ):
        # when
        task_forward_notification_to_discord(
            notification_id=1,
            discord_uid=2,
            title="title",
            message="message",
            level="INFO",
            timestamp="abc",
        )
        # then
        self.assertTrue(mock_forward_notification_to_discord.called)
        self.assertTrue(mock_logger.info.called)

    def test_should_retry_when_timeout_reached(
        self, mock_forward_notification_to_discord, mock_logger
    ):
        # given
        timeout_exception = DiscordProxyTimeoutError(
            status=grpc.StatusCode.DEADLINE_EXCEEDED, details="Timeout"
        )
        mock_forward_notification_to_discord.side_effect = timeout_exception
        # when
        with self.assertRaises(Retry):
            task_forward_notification_to_discord(
                notification_id=1,
                discord_uid=2,
                title="title",
                message="message",
                level="INFO",
                timestamp="abc",
            )
        self.assertTrue(mock_forward_notification_to_discord.called)
        self.assertTrue(mock_logger.warning.called)



"""Routes for Discord Notify."""

from django.urls import path

from . import views

app_name = "discordnotify"

urlpatterns = [
    path("test", views.send_test_notification, name="send_test_notification"),
]



from django.apps import AppConfig


class DiscordnotifyConfig(AppConfig):
    name = "discordnotify"
    label = "discordnotify"
    verbose_name = "discordnotify"

    def ready(self):
        from . import signals  # noqa F401



"""Core logic for Discord Notify."""

from typing import Any

from discordproxy.client import DiscordClient
from discordproxy.discord_api_pb2 import Embed

from allianceauth.notifications.models import Notification
from app_utils.urls import reverse_absolute, static_file_absolute_url

from . import __title__
from .app_settings import (
    DISCORDNOTIFY_MARK_AS_VIEWED,
    DISCORDPROXY_HOST,
    DISCORDPROXY_PORT,
)

# embed colors
COLOR_INFO = 0x5BC0DE
COLOR_SUCCESS = 0x5CB85C
COLOR_WARNING = 0xF0AD4E
COLOR_DANGER = 0xD9534F

COLOR_MAP = {
    "info": COLOR_INFO,
    "success": COLOR_SUCCESS,
    "warning": COLOR_WARNING,
    "danger": COLOR_DANGER,
}

# limits
MAX_LENGTH_TITLE = 256
MAX_LENGTH_DESCRIPTION = 2048
DISCORD_PROXY_TIMEOUT = 300  # in seconds


def forward_notification_to_discord(
    notification_id: int,
    discord_uid: int,
    title: str,
    message: str,
    level: str,
    timestamp: str,
):
    """Forward a new notification to Discord."""
    embed = _build_embed(
        notification_id=notification_id,
        title=title,
        message=message,
        level=level,
        timestamp=timestamp,
    )
    target = f"{DISCORDPROXY_HOST}:{DISCORDPROXY_PORT}"
    client = DiscordClient(target=target, timeout=DISCORD_PROXY_TIMEOUT)
    client.create_direct_message(user_id=discord_uid, embed=embed)
    _mark_as_viewed(notification_id)


def _build_embed(
    notification_id: int, title: str, message: str, level: str, timestamp: str
) -> Any:
    description = message.strip()
    if len(description) > MAX_LENGTH_DESCRIPTION:
        description = description[: (MAX_LENGTH_DESCRIPTION - 6)] + " [...]"
    author = Embed.Author(
        name="Alliance Auth Notification",
        icon_url=static_file_absolute_url("icons/apple-touch-icon.png"),
    )
    footer = Embed.Footer(
        text=__title__,
        icon_url=static_file_absolute_url("discordnotify/discordnotify_logo.png"),
    )
    return Embed(
        author=author,
        title=title.strip()[:MAX_LENGTH_TITLE],
        url=reverse_absolute("notifications:view", args=[notification_id]),
        description=description,
        color=COLOR_MAP.get(level, None),
        timestamp=timestamp,
        footer=footer,
    )


def _mark_as_viewed(notification_id):
    if DISCORDNOTIFY_MARK_AS_VIEWED:
        try:
            notif = Notification.objects.get(id=notification_id)
        except Notification.DoesNotExist:
            return
        notif.mark_viewed()



"""Settings for Discord Notify."""

from django.conf import settings

DISCORDPROXY_PORT = getattr(settings, "DISCORDPROXY_PORT", 50051)
"""Port used to communicate with Discord Proxy."""

DISCORDPROXY_HOST = getattr(settings, "DISCORDPROXY_HOST", "localhost")
"""Host used to communicate with Discord Proxy."""

DISCORDNOTIFY_SUPERUSER_ONLY = getattr(settings, "DISCORDNOTIFY_SUPERUSER_ONLY", False)
"""When set to True, only superusers will be get their notifications forwarded."""

DISCORDNOTIFY_ENABLED = getattr(settings, "DISCORDNOTIFY_ENABLED", True)
"""Set to False to disable this app."""

DISCORDNOTIFY_MARK_AS_VIEWED = getattr(settings, "DISCORDNOTIFY_MARK_AS_VIEWED", True)
"""When set to True, will mark all notifications as read
that have been successfully submitted to Discord.
"""
